import numpy as np
import pandas as pd


def codes_for_resampling(group_as_index: pd.Index, freq: str) -> np.ndarray:
    s = pd.Series(np.arange(group_as_index.size), group_as_index)
    grouped = s.groupby(pd.Grouper(freq=freq))
    first_items = grouped.first()
    counts = grouped.count()
    codes = np.repeat(np.arange(len(first_items)), counts)
    return codes



from __future__ import annotations

import copy
import logging
import warnings
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from functools import cached_property, partial
from typing import TYPE_CHECKING, Any, Literal, TypedDict

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike, DTypeLike

from . import aggregate_flox, aggregate_npg, xrutils
from . import xrdtypes as dtypes

if TYPE_CHECKING:
    FuncTuple = tuple[Callable | str, ...]
    OptionalFuncTuple = tuple[Callable | str | None, ...]


logger = logging.getLogger("flox")
T_ScanBinaryOpMode = Literal["apply_binary_op", "concat_then_scan"]


def _is_arg_reduction(func: str | Aggregation) -> bool:
    if isinstance(func, str) and func in ["argmin", "argmax", "nanargmax", "nanargmin"]:
        return True
    if isinstance(func, Aggregation) and func.reduction_type == "argreduce":
        return True
    return False


class AggDtypeInit(TypedDict):
    final: DTypeLike | None
    intermediate: tuple[DTypeLike, ...]


class AggDtype(TypedDict):
    user: DTypeLike | None
    final: np.dtype
    numpy: tuple[np.dtype | type[np.intp], ...]
    intermediate: tuple[np.dtype | type[np.intp], ...]


def get_npg_aggregation(func, *, engine):
    try:
        method_ = getattr(aggregate_npg, func)
        method = partial(method_, engine=engine)
    except AttributeError:
        aggregate = aggregate_npg._get_aggregate(engine).aggregate
        method = partial(aggregate, func=func)
    return method


def generic_aggregate(
    group_idx,
    array,
    *,
    engine: str,
    func: str,
    axis=-1,
    size=None,
    fill_value=None,
    dtype=None,
    **kwargs,
):
    if func == "identity":
        return array

    if func in ["nanfirst", "nanlast"] and array.dtype.kind in "US":
        func = func[3:]

    if engine == "flox":
        try:
            method = getattr(aggregate_flox, func)
        except AttributeError:
            # logger.debug(f"Couldn't find {func} for engine='flox'. Falling back to numpy")
            method = get_npg_aggregation(func, engine="numpy")

    elif engine == "numbagg":
        from . import aggregate_numbagg

        try:
            if "var" in func or "std" in func:
                ddof = kwargs.get("ddof", 0)
                if aggregate_numbagg.NUMBAGG_SUPPORTS_DDOF or (ddof != 0):
                    method = getattr(aggregate_numbagg, func)
                else:
                    logger.debug(f"numbagg too old for ddof={ddof}. Falling back to numpy")
                    method = get_npg_aggregation(func, engine="numpy")
            else:
                method = getattr(aggregate_numbagg, func)

        except AttributeError:
            # logger.debug(f"Couldn't find {func} for engine='numbagg'. Falling back to numpy")
            method = get_npg_aggregation(func, engine="numpy")

    elif engine in ["numpy", "numba"]:
        method = get_npg_aggregation(func, engine=engine)

    else:
        raise ValueError(
            f"Expected engine to be one of ['flox', 'numpy', 'numba', 'numbagg']. Received {engine} instead."
        )

    group_idx = np.asarray(group_idx, like=array)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", r"All-NaN (slice|axis) encountered")
        result = method(
            group_idx,
            array,
            axis=axis,
            size=size,
            fill_value=fill_value,
            dtype=dtype,
            **kwargs,
        )
    return result


def _atleast_1d(inp, min_length: int = 1):
    if xrutils.is_scalar(inp):
        inp = (inp,) * min_length
    assert len(inp) >= min_length
    return inp


def returns_empty_tuple(*args, **kwargs):
    return ()


@dataclass
class Dim:
    values: ArrayLike
    name: str | None

    @cached_property
    def is_scalar(self) -> bool:
        return xrutils.is_scalar(self.values)

    @cached_property
    def size(self) -> int:
        return 0 if self.is_scalar else len(self.values)  # type: ignore[arg-type]


class Aggregation:
    def __init__(
        self,
        name: str,
        *,
        numpy: str | None = None,
        chunk: str | FuncTuple | None,
        combine: str | FuncTuple | None,
        preprocess: Callable | None = None,
        finalize: Callable | None = None,
        fill_value=None,
        final_fill_value=dtypes.NA,
        dtypes=None,
        final_dtype: DTypeLike | None = None,
        reduction_type: Literal["reduce", "argreduce"] = "reduce",
        new_dims_func: Callable | None = None,
        preserves_dtype: bool = False,
    ):
        """
        Blueprint for computing grouped aggregations.

        See aggregations.py for examples on how to specify reductions.

        Attributes
        ----------
        name : str
            Name of reduction.
        numpy : str or callable, optional
            Reduction function applied to numpy inputs. This function should
            compute the grouped reduction and must have a specific signature.
            If string, these must be "native" reductions implemented by the backend
            engines (numpy_groupies, flox, numbagg). If None, will be set to ``name``.
        chunk : None or str or tuple of str or callable or tuple of callable
            For dask inputs only. Either a single function or a list of
            functions to be applied blockwise on the input dask array. If None, will raise
            an error for dask inputs.
        combine : None or str or tuple of str or callbe or tuple of callable
            For dask inputs only. Functions applied when combining intermediate
            results from the blockwise stage (see ``chunk``). If None, will raise an error
            for dask inputs.
        finalize : callable
            For dask inputs only. Function that combines intermediate results to compute
            final result.
        preprocess : callable
            For dask inputs only. Preprocess inputs before ``chunk`` stage.
        reduction_type : {"reduce", "argreduce"}
            Type of reduction.
        fill_value : number or tuple(number), optional
            Value to use when a group has no members. If single value will be converted
            to tuple of same length as chunk. If appropriate, provide a different fill_value
            per reduction in ``chunk`` as a tuple.
        final_fill_value : optional
            fill_value for final result.
        dtypes : DType or tuple(DType), optional
            dtypes for intermediate results. If single value, will be converted to a tuple
            of same length as chunk. If appropriate, provide a different fill_value
            per reduction in ``chunk`` as a tuple.
        final_dtype : DType, optional
            DType for output. By default, uses dtype of array being reduced.
        new_dims_func: Callable
            Function that receives finalize_kwargs and returns a tupleof sizes of any new dimensions
            added by the reduction. For e.g. quantile for q=(0.5, 0.85) adds a new dimension of size 2,
            so returns (2,)
        preserves_dtype: bool,
           Whether a function preserves the dtype on return E.g. min, max, first, last, mode
        """
        self.name = name
        # preprocess before blockwise
        self.preprocess = preprocess
        # Use "chunk_reduce" or "chunk_argreduce"
        self.reduction_type = reduction_type
        self.numpy: FuncTuple = (numpy,) if numpy is not None else (self.name,)
        # initialize blockwise reduction
        self.chunk: OptionalFuncTuple = _atleast_1d(chunk)
        # how to aggregate results after first round of reduction
        self.combine: OptionalFuncTuple = _atleast_1d(combine)
        # simpler reductions used with the "simple combine" algorithm
        self.simple_combine: OptionalFuncTuple = ()
        # finalize results (see mean)
        self.finalize: Callable | None = finalize

        self.fill_value = {}
        # This is used for the final reindexing
        self.fill_value[name] = final_fill_value
        # Aggregation.fill_value is used to reindex to group labels
        # at the *intermediate* step.
        # They should make sense when aggregated together with results from other blocks
        self.fill_value["intermediate"] = self._normalize_dtype_fill_value(fill_value, "fill_value")

        self.dtype_init: AggDtypeInit = {
            "final": final_dtype,
            "intermediate": self._normalize_dtype_fill_value(dtypes, "dtype"),
        }
        self.dtype: AggDtype = None  # type: ignore[assignment]

        # The following are set by _initialize_aggregation
        self.finalize_kwargs: dict[Any, Any] = {}
        self.min_count: int = 0
        self.new_dims_func: Callable = returns_empty_tuple if new_dims_func is None else new_dims_func
        self.preserves_dtype = preserves_dtype

    @cached_property
    def new_dims(self) -> tuple[Dim]:
        return self.new_dims_func(**self.finalize_kwargs)

    @cached_property
    def num_new_vector_dims(self) -> int:
        return len(tuple(dim for dim in self.new_dims if not dim.is_scalar))

    def _normalize_dtype_fill_value(self, value, name):
        value = _atleast_1d(value)
        if len(value) == 1 and len(value) < len(self.chunk):
            value = value * len(self.chunk)
        if len(value) != len(self.chunk):
            raise ValueError(f"Bad {name} specified for Aggregation {name}.")
        return value

    def __dask_tokenize__(self):
        return (
            Aggregation,
            self.name,
            self.preprocess,
            self.reduction_type,
            self.numpy,
            self.chunk,
            self.combine,
            self.finalize,
            self.fill_value,
            self.dtype,
        )

    def __repr__(self) -> str:
        return "\n".join(
            (
                f"{self.name!r}, fill: {self.fill_value.values()!r}, dtype: {self.dtype}",
                f"chunk: {self.chunk!r}",
                f"combine: {self.combine!r}",
                f"finalize: {self.finalize!r}",
                f"min_count: {self.min_count!r}",
            )
        )


count = Aggregation(
    "count",
    numpy="nanlen",
    chunk="nanlen",
    combine="sum",
    fill_value=0,
    final_fill_value=0,
    dtypes=np.intp,
    final_dtype=np.intp,
)

# note that the fill values are the result of np.func([np.nan, np.nan])
# final_fill_value is used for groups that don't exist. This is usually np.nan
sum_ = Aggregation("sum", chunk="sum", combine="sum", fill_value=0)
nansum = Aggregation("nansum", chunk="nansum", combine="sum", fill_value=0)
prod = Aggregation("prod", chunk="prod", combine="prod", fill_value=1, final_fill_value=1)
nanprod = Aggregation("nanprod", chunk="nanprod", combine="prod", fill_value=1)


def _mean_finalize(sum_, count):
    with np.errstate(invalid="ignore", divide="ignore"):
        return sum_ / count


mean = Aggregation(
    "mean",
    chunk=("sum", "nanlen"),
    combine=("sum", "sum"),
    finalize=_mean_finalize,
    fill_value=(0, 0),
    dtypes=(None, np.intp),
    final_dtype=np.floating,
)
nanmean = Aggregation(
    "nanmean",
    chunk=("nansum", "nanlen"),
    combine=("sum", "sum"),
    finalize=_mean_finalize,
    fill_value=(0, 0),
    dtypes=(None, np.intp),
    final_dtype=np.floating,
)


# TODO: fix this for complex numbers
def _var_finalize(sumsq, sum_, count, ddof=0):
    with np.errstate(invalid="ignore", divide="ignore"):
        result = (sumsq - (sum_**2 / count)) / (count - ddof)
    result[count <= ddof] = np.nan
    return result


def _std_finalize(sumsq, sum_, count, ddof=0):
    return np.sqrt(_var_finalize(sumsq, sum_, count, ddof))


# var, std always promote to float, so we set nan
var = Aggregation(
    "var",
    chunk=("sum_of_squares", "sum", "nanlen"),
    combine=("sum", "sum", "sum"),
    finalize=_var_finalize,
    fill_value=0,
    final_fill_value=np.nan,
    dtypes=(None, None, np.intp),
    final_dtype=np.floating,
)
nanvar = Aggregation(
    "nanvar",
    chunk=("nansum_of_squares", "nansum", "nanlen"),
    combine=("sum", "sum", "sum"),
    finalize=_var_finalize,
    fill_value=0,
    final_fill_value=np.nan,
    dtypes=(None, None, np.intp),
    final_dtype=np.floating,
)
std = Aggregation(
    "std",
    chunk=("sum_of_squares", "sum", "nanlen"),
    combine=("sum", "sum", "sum"),
    finalize=_std_finalize,
    fill_value=0,
    final_fill_value=np.nan,
    dtypes=(None, None, np.intp),
    final_dtype=np.floating,
)
nanstd = Aggregation(
    "nanstd",
    chunk=("nansum_of_squares", "nansum", "nanlen"),
    combine=("sum", "sum", "sum"),
    finalize=_std_finalize,
    fill_value=0,
    final_fill_value=np.nan,
    dtypes=(None, None, np.intp),
    final_dtype=np.floating,
)


min_ = Aggregation("min", chunk="min", combine="min", fill_value=dtypes.INF, preserves_dtype=True)
nanmin = Aggregation(
    "nanmin",
    chunk="nanmin",
    combine="nanmin",
    fill_value=dtypes.INF,
    final_fill_value=dtypes.NA,
    preserves_dtype=True,
)
max_ = Aggregation("max", chunk="max", combine="max", fill_value=dtypes.NINF, preserves_dtype=True)
nanmax = Aggregation(
    "nanmax",
    chunk="nanmax",
    combine="nanmax",
    fill_value=dtypes.NINF,
    final_fill_value=dtypes.NA,
    preserves_dtype=True,
)


def argreduce_preprocess(array, axis):
    """Returns a tuple of array, index along axis.

    Copied from dask.array.chunk.argtopk_preprocess
    """
    import dask.array
    import numpy as np

    # TODO: arg reductions along multiple axes seems weird.
    assert len(axis) == 1
    axis = axis[0]

    idx = dask.array.arange(array.shape[axis], chunks=array.chunks[axis], dtype=np.intp)
    # broadcast (TODO: is this needed?)
    idx = idx[tuple(slice(None) if i == axis else np.newaxis for i in range(array.ndim))]

    def _zip_index(array_, idx_):
        return (array_, idx_)

    return dask.array.map_blocks(
        _zip_index,
        array,
        idx,
        dtype=array.dtype,
        meta=array._meta,
        name="groupby-argreduce-preprocess",
    )


def _pick_second(*x):
    return x[1]


argmax = Aggregation(
    "argmax",
    preprocess=argreduce_preprocess,
    chunk=("max", "argmax"),  # order is important
    combine=("max", "argmax"),
    reduction_type="argreduce",
    fill_value=(dtypes.NINF, 0),
    final_fill_value=-1,
    finalize=_pick_second,
    dtypes=(None, np.intp),
    final_dtype=np.intp,
)

argmin = Aggregation(
    "argmin",
    preprocess=argreduce_preprocess,
    chunk=("min", "argmin"),  # order is important
    combine=("min", "argmin"),
    reduction_type="argreduce",
    fill_value=(dtypes.INF, 0),
    final_fill_value=-1,
    finalize=_pick_second,
    dtypes=(None, np.intp),
    final_dtype=np.intp,
)

nanargmax = Aggregation(
    "nanargmax",
    preprocess=argreduce_preprocess,
    chunk=("nanmax", "nanargmax"),  # order is important
    combine=("max", "argmax"),
    reduction_type="argreduce",
    fill_value=(dtypes.NINF, 0),
    final_fill_value=-1,
    finalize=_pick_second,
    dtypes=(None, np.intp),
    final_dtype=np.intp,
)

nanargmin = Aggregation(
    "nanargmin",
    preprocess=argreduce_preprocess,
    chunk=("nanmin", "nanargmin"),  # order is important
    combine=("min", "argmin"),
    reduction_type="argreduce",
    fill_value=(dtypes.INF, 0),
    final_fill_value=-1,
    finalize=_pick_second,
    dtypes=(None, np.intp),
    final_dtype=np.intp,
)

first = Aggregation("first", chunk=None, combine=None, fill_value=None, preserves_dtype=True)
last = Aggregation("last", chunk=None, combine=None, fill_value=None, preserves_dtype=True)
nanfirst = Aggregation(
    "nanfirst",
    chunk="nanfirst",
    combine="nanfirst",
    fill_value=dtypes.NA,
    preserves_dtype=True,
)
nanlast = Aggregation(
    "nanlast",
    chunk="nanlast",
    combine="nanlast",
    fill_value=dtypes.NA,
    preserves_dtype=True,
)

all_ = Aggregation(
    "all",
    chunk="all",
    combine="all",
    fill_value=True,
    final_fill_value=False,
    dtypes=bool,
    final_dtype=bool,
)
any_ = Aggregation(
    "any",
    chunk="any",
    combine="any",
    fill_value=False,
    final_fill_value=False,
    dtypes=bool,
    final_dtype=bool,
)

# Support statistical quantities only blockwise
# The parallel versions will be approximate and are hard to implement!
median = Aggregation(
    name="median",
    fill_value=dtypes.NA,
    chunk=None,
    combine=None,
    final_dtype=np.floating,
)
nanmedian = Aggregation(
    name="nanmedian",
    fill_value=dtypes.NA,
    chunk=None,
    combine=None,
    final_dtype=np.floating,
)


def quantile_new_dims_func(q) -> tuple[Dim]:
    return (Dim(name="quantile", values=q),)


# if the input contains integers or floats smaller than float64,
# the output data-type is float64. Otherwise, the output data-type is the same as that
# of the input.
quantile = Aggregation(
    name="quantile",
    fill_value=dtypes.NA,
    chunk=None,
    combine=None,
    final_dtype=np.float64,
    new_dims_func=quantile_new_dims_func,
)
nanquantile = Aggregation(
    name="nanquantile",
    fill_value=dtypes.NA,
    chunk=None,
    combine=None,
    final_dtype=np.float64,
    new_dims_func=quantile_new_dims_func,
)
mode = Aggregation(name="mode", fill_value=dtypes.NA, chunk=None, combine=None, preserves_dtype=True)
nanmode = Aggregation(name="nanmode", fill_value=dtypes.NA, chunk=None, combine=None, preserves_dtype=True)


@dataclass
class Scan:
    # This dataclass is separate from Aggregations since there's not much in common
    # between reductions and scans
    name: str
    # binary operation (e.g. np.add)
    # Must be None for mode="concat_then_scan"
    binary_op: Callable | None
    # in-memory grouped scan function (e.g. cumsum)
    scan: str
    # Grouped reduction that yields the last result of the scan (e.g. sum)
    reduction: str
    # Identity element
    identity: Any
    # dtype of result
    dtype: Any = None
    # "Mode" of applying binary op.
    # for np.add we apply the op directly to the `state` array and the `current` array.
    # for ffill, bfill we concat `state` to `current` and then run the scan again.
    mode: T_ScanBinaryOpMode = "apply_binary_op"
    preprocess: Callable | None = None
    finalize: Callable | None = None


def concatenate(arrays: Sequence[AlignedArrays], axis=-1, out=None) -> AlignedArrays:
    group_idx = np.concatenate([a.group_idx for a in arrays], axis=axis)
    array = np.concatenate([a.array for a in arrays], axis=axis)
    return AlignedArrays(array=array, group_idx=group_idx)


@dataclass
class AlignedArrays:
    """Simple Xarray DataArray type data class with two aligned arrays."""

    array: np.ndarray
    group_idx: np.ndarray

    def __post_init__(self):
        assert self.array.shape[-1] == self.group_idx.size

    def last(self) -> AlignedArrays:
        from flox.core import chunk_reduce

        reduced = chunk_reduce(
            self.array,
            self.group_idx,
            func=("nanlast",),
            axis=-1,
            # TODO: automate?
            engine="flox",
            dtype=self.array.dtype,
            fill_value=dtypes._get_fill_value(self.array.dtype, dtypes.NA),
            expected_groups=None,
        )
        return AlignedArrays(array=reduced["intermediates"][0], group_idx=reduced["groups"])


@dataclass
class ScanState:
    """Dataclass representing intermediates for scan."""

    # last value of each group seen so far
    state: AlignedArrays | None
    # intermediate result
    result: AlignedArrays | None

    def __post_init__(self):
        assert (self.state is not None) or (self.result is not None)


def reverse(a: AlignedArrays) -> AlignedArrays:
    a.group_idx = a.group_idx[..., ::-1]
    a.array = a.array[..., ::-1]
    return a


def scan_binary_op(left_state: ScanState, right_state: ScanState, *, agg: Scan) -> ScanState:
    from .core import reindex_

    assert left_state.state is not None
    left = left_state.state
    right = right_state.result if right_state.result is not None else right_state.state
    assert right is not None

    if agg.mode == "apply_binary_op":
        assert agg.binary_op is not None
        # Implements groupby binary operation.
        reindexed = reindex_(
            left.array,
            from_=pd.Index(left.group_idx),
            # can't use right.group_idx since we need to do the indexing later
            to=pd.RangeIndex(right.group_idx.max() + 1),
            fill_value=agg.identity,
            axis=-1,
        )
        result = AlignedArrays(
            array=agg.binary_op(reindexed[..., right.group_idx], right.array),
            group_idx=right.group_idx,
        )

    elif agg.mode == "concat_then_scan":
        # Implements the binary op portion of the scan as a concatenate-then-scan.
        # This is useful for `ffill`, and presumably more generalized scans.
        assert agg.binary_op is None
        concat = concatenate([left, right], axis=-1)
        final_value = generic_aggregate(
            concat.group_idx,
            concat.array,
            func=agg.scan,
            axis=concat.array.ndim - 1,
            engine="flox",
            fill_value=agg.identity,
        )
        result = AlignedArrays(array=final_value[..., left.group_idx.size :], group_idx=right.group_idx)
    else:
        raise ValueError(f"Unknown binary op application mode: {agg.mode!r}")

    # This is quite important. We need to update the state seen so far and propagate that.
    # So we must account for what we know when entering this function: i.e. `left`
    # TODO: this is a bit wasteful since it will sort again, but for now let's focus on
    # correctness and DRY
    lasts = concatenate([left, result]).last()

    return ScanState(
        state=lasts,
        # The binary op is called on the results of the reduction too when building up the tree.
        # We need to be careful and assign those results only to `state` and not the final result.
        # Up above, `result` is privileged when it exists.
        result=None if right_state.result is None else result,
    )


# TODO: numpy_groupies cumsum is a broken when NaNs are present.
# cumsum = Scan("cumsum", binary_op=np.add, reduction="sum", scan="cumsum", identity=0)
nancumsum = Scan("nancumsum", binary_op=np.add, reduction="nansum", scan="nancumsum", identity=0)
# ffill uses the identity for scan, and then at the binary-op state,
# we concatenate the blockwise-reduced values with the original block,
# and then execute the scan
# TODO: consider adding chunk="identity" here, like with reductions as an optimization
ffill = Scan(
    "ffill",
    binary_op=None,
    reduction="nanlast",
    scan="ffill",
    # Important: this must be NaN otherwise, ffill does not work.
    identity=np.nan,
    mode="concat_then_scan",
)
bfill = Scan(
    "bfill",
    binary_op=None,
    reduction="nanlast",
    scan="ffill",
    # Important: this must be NaN otherwise, bfill does not work.
    identity=np.nan,
    mode="concat_then_scan",
    preprocess=reverse,
    finalize=reverse,
)
# TODO: not implemented in numpy_groupies
# cumprod = Scan("cumprod", binary_op=np.multiply, preop="prod", scan="cumprod")


AGGREGATIONS: dict[str, Aggregation | Scan] = {
    "any": any_,
    "all": all_,
    "count": count,
    "sum": sum_,
    "nansum": nansum,
    "prod": prod,
    "nanprod": nanprod,
    "mean": mean,
    "nanmean": nanmean,
    "var": var,
    "nanvar": nanvar,
    "std": std,
    "nanstd": nanstd,
    "max": max_,
    "nanmax": nanmax,
    "min": min_,
    "nanmin": nanmin,
    "argmax": argmax,
    "nanargmax": nanargmax,
    "argmin": argmin,
    "nanargmin": nanargmin,
    "first": first,
    "nanfirst": nanfirst,
    "last": last,
    "nanlast": nanlast,
    "median": median,
    "nanmedian": nanmedian,
    "quantile": quantile,
    "nanquantile": nanquantile,
    "mode": mode,
    "nanmode": nanmode,
    # "cumsum": cumsum,
    "nancumsum": nancumsum,
    "ffill": ffill,
    "bfill": bfill,
}


def _initialize_aggregation(
    func: str | Aggregation,
    dtype,
    array_dtype,
    fill_value,
    min_count: int,
    finalize_kwargs: dict[Any, Any] | None,
) -> Aggregation:
    agg: Aggregation
    if not isinstance(func, Aggregation):
        try:
            # TODO: need better interface
            # we set dtype, fillvalue on reduction later. so deepcopy now
            agg_ = copy.deepcopy(AGGREGATIONS[func])
            assert isinstance(agg_, Aggregation)
            agg = agg_
        except KeyError:
            raise NotImplementedError(f"Reduction {func!r} not implemented yet")
    elif isinstance(func, Aggregation):
        # TODO: test that func is a valid Aggregation
        agg = copy.deepcopy(func)
        func = agg.name
    else:
        raise ValueError("Bad type for func. Expected str or Aggregation")

    # np.dtype(None) == np.dtype("float64")!!!
    # so check for not None
    dtype_: np.dtype | None = (
        np.dtype(dtype) if dtype is not None and not isinstance(dtype, np.dtype) else dtype
    )
    final_dtype = dtypes._normalize_dtype(
        dtype_ or agg.dtype_init["final"], array_dtype, agg.preserves_dtype, fill_value
    )
    agg.dtype = {
        "user": dtype,  # Save to automatically choose an engine
        "final": final_dtype,
        "numpy": (final_dtype,),
        "intermediate": tuple(
            (
                dtypes._normalize_dtype(int_dtype, np.result_type(array_dtype, final_dtype), int_fv)
                if int_dtype is None
                else np.dtype(int_dtype)
            )
            for int_dtype, int_fv in zip(agg.dtype_init["intermediate"], agg.fill_value["intermediate"])
        ),
    }

    # Replace sentinel fill values according to dtype
    agg.fill_value["user"] = fill_value
    agg.fill_value["intermediate"] = tuple(
        dtypes._get_fill_value(dt, fv)
        for dt, fv in zip(agg.dtype["intermediate"], agg.fill_value["intermediate"])
    )
    agg.fill_value[func] = dtypes._get_fill_value(agg.dtype["final"], agg.fill_value[func])

    fv = fill_value if fill_value is not None else agg.fill_value[agg.name]
    if _is_arg_reduction(agg):
        # this allows us to unravel_index easily. we have to do that nearly every time.
        agg.fill_value["numpy"] = (0,)
    else:
        agg.fill_value["numpy"] = (fv,)

    if finalize_kwargs is not None:
        assert isinstance(finalize_kwargs, dict)
        agg.finalize_kwargs = finalize_kwargs

    # This is needed for the dask pathway.
    # Because we use intermediate fill_value since a group could be
    # absent in one block, but present in another block
    # We set it for numpy to get nansum, nanprod tests to pass
    # where the identity element is 0, 1
    # Also needed for nanmin, nanmax where intermediate fill_value is +-np.inf,
    # but final_fill_value is dtypes.NA
    if (
        # TODO: this is a total hack, setting a default fill_value
        # even though numpy doesn't define identity for nanmin, nanmax
        agg.name in ["nanmin", "nanmax"] and min_count == 0
    ):
        min_count = 1
        agg.fill_value["user"] = agg.fill_value["user"] or agg.fill_value[agg.name]

    if min_count > 0:
        agg.min_count = min_count
        agg.numpy += ("nanlen",)
        if agg.chunk != (None,):
            agg.chunk += ("nanlen",)
            agg.combine += ("sum",)
        agg.fill_value["intermediate"] += (0,)
        agg.fill_value["numpy"] += (0,)
        agg.dtype["intermediate"] += (np.intp,)
        agg.dtype["numpy"] += (np.intp,)
    else:
        agg.min_count = 0

    simple_combine: list[Callable | None] = []
    for combine in agg.combine:
        if isinstance(combine, str):
            if combine in ["nanfirst", "nanlast"]:
                simple_combine.append(getattr(xrutils, combine))
            else:
                simple_combine.append(getattr(np, combine))
        else:
            simple_combine.append(combine)

    agg.simple_combine = tuple(simple_combine)

    return agg



import warnings
from collections.abc import Callable
from typing import Any

import pandas as pd
import pytest

pytest.importorskip("hypothesis")
pytest.importorskip("dask")
pytest.importorskip("cftime")

import dask
import hypothesis.extra.numpy as npst
import hypothesis.strategies as st
import numpy as np
from hypothesis import assume, given, note, settings

import flox
from flox.core import groupby_reduce, groupby_scan
from flox.xrutils import _contains_cftime_datetimes, _to_pytimedelta, datetime_to_numeric, isnull, notnull

from . import BLOCKWISE_FUNCS, assert_equal
from .strategies import all_arrays, by_arrays, chunked_arrays, func_st, numeric_dtypes, numeric_like_arrays
from .strategies import chunks as chunks_strategy

dask.config.set(scheduler="sync")


def ffill(array, axis, dtype=None):
    return flox.aggregate_flox.ffill(np.zeros(array.shape[-1], dtype=int), array, axis=axis)


def bfill(array, axis, dtype=None):
    return flox.aggregate_flox.ffill(
        np.zeros(array.shape[-1], dtype=int),
        array[::-1],
        axis=axis,
    )[::-1]


NUMPY_SCAN_FUNCS: dict[str, Callable] = {
    "nancumsum": np.nancumsum,
    "ffill": ffill,
    "bfill": bfill,
}  # "cumsum": np.cumsum,


def not_overflowing_array(array: np.ndarray[Any, Any]) -> bool:
    if array.dtype.kind in "Mm":
        array = array.view(np.int64)
    if array.dtype.kind == "f":
        info = np.finfo(array.dtype)
    elif array.dtype.kind in ["i", "u"]:
        info = np.iinfo(array.dtype)  # type: ignore[assignment]
    else:
        return True

    array = array.ravel()
    array = array[notnull(array)]
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        result = bool(np.all((array < info.max / array.size) & (array > info.min / array.size)))
    # note(f"returning {result}, {array.min()} vs {info.min}, {array.max()} vs {info.max}")
    return result


@given(
    data=st.data(),
    array=st.one_of(all_arrays, chunked_arrays()),
    func=func_st,
)
@settings(deadline=None)
def test_groupby_reduce(data, array, func: str) -> None:
    # overflow behaviour differs between bincount and sum (for example)
    assume(not_overflowing_array(array))
    # TODO: fix var for complex numbers upstream
    assume(not (("quantile" in func or "var" in func or "std" in func) and array.dtype.kind == "c"))
    assume(not ("quantile" in func and array.dtype.kind == "b"))
    # arg* with nans in array are weird
    assume("arg" not in func and not np.any(isnull(array).ravel()))

    # TODO: funny bugs with overflows here
    is_cftime = _contains_cftime_datetimes(array)
    assume(not (is_cftime and func in ["prod", "nanprod"]))

    axis = -1
    by = data.draw(
        by_arrays(
            elements={
                "alphabet": st.just("a"),
                "min_value": 1,
                "max_value": 1,
                "min_size": 1,
                "max_size": 1,
            },
            shape=st.just((array.shape[-1],)),
        )
    )
    if func in BLOCKWISE_FUNCS and isinstance(array, dask.array.Array):
        array = array.rechunk({axis: -1})
    assert len(np.unique(by)) == 1
    kwargs = {"q": 0.8} if "quantile" in func else {}
    flox_kwargs: dict[str, Any] = {}
    with np.errstate(invalid="ignore", divide="ignore"):
        actual, *_ = groupby_reduce(
            array,
            by,
            func=func,
            axis=axis,
            engine="numpy",
            **flox_kwargs,
            finalize_kwargs=kwargs,
        )

        # numpy-groupies always does the calculation in float64
        if (
            ("var" in func or "std" in func or "sum" in func or "mean" in func or "quantile" in func)
            and array.dtype.kind == "f"
            and array.dtype.itemsize != 8
        ):
            # bincount always accumulates in float64,
            # casting to float64 handles std more like npg does.
            # Setting dtype=float64 works fine for sum, mean.
            cast_to = array.dtype
            array = array.astype(np.float64)
            note(f"casting array to float64, cast_to={cast_to!r}")
        else:
            cast_to = None

        if array.dtype.kind in "Mm":
            array = array.view(np.int64)
            cast_to = array.dtype
        elif is_cftime:
            offset = array.min()
            array = datetime_to_numeric(array, offset, datetime_unit="us")
        note(("kwargs:", kwargs, "cast_to:", cast_to))
        expected = getattr(np, func)(array, axis=axis, keepdims=True, **kwargs)
        if cast_to is not None:
            note(("casting to:", cast_to))
            expected = expected.astype(cast_to)
            actual = actual.astype(cast_to)
        if is_cftime:
            expected = _to_pytimedelta(expected, unit="us") + offset

    note(("expected: ", expected, "actual: ", actual))
    tolerance = {"atol": 1e-15}
    assert_equal(expected, actual, tolerance)


@given(
    data=st.data(),
    array=chunked_arrays(arrays=numeric_like_arrays),
    func=func_st,
)
def test_groupby_reduce_numpy_vs_dask(data, array, func: str) -> None:
    numpy_array = array.compute()
    # overflow behaviour differs between bincount and sum (for example)
    assume(not_overflowing_array(numpy_array))
    # TODO: fix var for complex numbers upstream
    assume(not (("quantile" in func or "var" in func or "std" in func) and array.dtype.kind == "c"))
    # # arg* with nans in array are weird
    assume("arg" not in func and not np.any(isnull(numpy_array.ravel())))
    if func in ["nanmedian", "nanquantile", "median", "quantile"]:
        array = array.rechunk({-1: -1})

    axis = -1
    by = data.draw(by_arrays(shape=st.just((array.shape[-1],))))
    kwargs = {"q": 0.8} if "quantile" in func else {}
    flox_kwargs: dict[str, Any] = {}

    kwargs = dict(
        func=func,
        axis=axis,
        engine="numpy",
        **flox_kwargs,
        finalize_kwargs=kwargs,
    )
    result_dask, *_ = groupby_reduce(array, by, **kwargs)
    result_numpy, *_ = groupby_reduce(numpy_array, by, **kwargs)
    assert_equal(result_numpy, result_dask)


@settings(report_multiple_bugs=False)
@given(
    data=st.data(),
    array=chunked_arrays(arrays=numeric_like_arrays),
    func=st.sampled_from(tuple(NUMPY_SCAN_FUNCS)),
)
def test_scans(data, array: dask.array.Array, func: str) -> None:
    if "cum" in func:
        assume(not_overflowing_array(np.asarray(array)))

    by = data.draw(by_arrays(shape=st.just((array.shape[-1],))))
    axis = array.ndim - 1

    # Too many float32 edge-cases!
    if "cum" in func and array.dtype.kind == "f" and array.dtype.itemsize == 4:
        assume(False)
    numpy_array = array.compute()
    if numpy_array.dtype.kind not in "Mm":
        assume((np.abs(numpy_array) < 2**53).all())

    if numpy_array.dtype.kind in "Mm":
        dtype = numpy_array.dtype
        asnumeric = numpy_array.view(np.int64)
    else:
        asnumeric = numpy_array
        dtype = NUMPY_SCAN_FUNCS[func](asnumeric[..., [0]], axis=axis).dtype
    expected = np.empty_like(numpy_array, dtype=dtype)
    group_idx, uniques = pd.factorize(by)
    for i in range(len(uniques)):
        mask = group_idx == i
        if not mask.any():
            note((by, group_idx, uniques))
            raise ValueError
        expected[..., mask] = NUMPY_SCAN_FUNCS[func](asnumeric[..., mask], axis=axis)

    if dtype:
        expected = expected.astype(dtype)
    note((numpy_array, group_idx, array.chunks))

    tolerance = {"rtol": 1e-13, "atol": 1e-15}
    actual = groupby_scan(numpy_array, by, func=func, axis=-1, dtype=dtype)
    assert_equal(actual, expected, tolerance)

    actual = groupby_scan(array, by, func=func, axis=-1, dtype=dtype)
    assert_equal(actual, expected, tolerance)


@given(data=st.data(), array=chunked_arrays())
def test_ffill_bfill_reverse(data, array: dask.array.Array) -> None:
    by = data.draw(by_arrays(shape=st.just((array.shape[-1],))))

    def reverse(arr):
        return arr[..., ::-1]

    forward = groupby_scan(array, by, func="ffill")
    as_numpy = groupby_scan(array.compute(), by, func="ffill")
    assert_equal(forward, as_numpy)

    backward = groupby_scan(array, by, func="bfill")
    as_numpy = groupby_scan(array.compute(), by, func="bfill")
    assert_equal(backward, as_numpy)

    backward_reversed = reverse(groupby_scan(reverse(array), reverse(by), func="bfill"))
    assert_equal(forward, backward_reversed)

    forward_reversed = reverse(groupby_scan(reverse(array), reverse(by), func="ffill"))
    assert_equal(forward_reversed, backward)


@given(
    data=st.data(),
    array=chunked_arrays(),
    func=st.sampled_from(["first", "last", "nanfirst", "nanlast"]),
)
def test_first_last(data, array: dask.array.Array, func: str) -> None:
    by = data.draw(by_arrays(shape=st.just((array.shape[-1],))))

    INVERSES = {
        "first": "last",
        "last": "first",
        "nanfirst": "nanlast",
        "nanlast": "nanfirst",
    }
    MATES = {
        "first": "nanfirst",
        "last": "nanlast",
        "nanfirst": "first",
        "nanlast": "last",
    }
    inverse = INVERSES[func]
    mate = MATES[func]

    if func in ["first", "last"]:
        array = array.rechunk((*array.chunks[:-1], -1))

    for arr in [array, array.compute()]:
        forward, *fg = groupby_reduce(arr, by, func=func, engine="flox")
        reverse, *rg = groupby_reduce(arr[..., ::-1], by[..., ::-1], func=inverse, engine="flox")

        assert forward.dtype == reverse.dtype
        assert forward.dtype == arr.dtype

        assert_equal(fg, rg)
        assert_equal(forward, reverse)

    if arr.dtype.kind == "f" and not isnull(array.compute()).any():
        if mate in ["first", "last"]:
            array = array.rechunk((*array.chunks[:-1], -1))

        first, *_ = groupby_reduce(array, by, func=func, engine="flox")
        second, *_ = groupby_reduce(array, by, func=mate, engine="flox")
        assert_equal(first, second)


@given(data=st.data(), func=st.sampled_from(["nanfirst", "nanlast"]))
def test_first_last_useless(data, func):
    shape = data.draw(npst.array_shapes())
    by = data.draw(by_arrays(shape=st.just(shape[slice(-1, None)])))
    chunks = data.draw(chunks_strategy(shape=shape))
    array = np.zeros(shape, dtype=np.int8)
    if chunks is not None:
        array = dask.array.from_array(array, chunks=chunks)
    actual, groups = groupby_reduce(array, by, axis=-1, func=func, engine="numpy")
    expected = np.zeros(shape[:-1] + (len(groups),), dtype=array.dtype)
    assert_equal(actual, expected)


@given(
    func=st.sampled_from(["sum", "prod", "nansum", "nanprod"]),
    engine=st.sampled_from(["numpy", "flox"]),
    array_dtype=st.none() | numeric_dtypes,
    dtype=st.none() | numeric_dtypes,
)
def test_agg_dtype_specified(func, array_dtype, dtype, engine):
    # regression test for GH388
    counts = np.array([0, 2, 1, 0, 1], dtype=array_dtype)
    group = np.array([1, 1, 1, 2, 2])
    actual, _ = groupby_reduce(
        counts,
        group,
        expected_groups=(np.array([1, 2]),),
        func=func,
        dtype=dtype,
        engine=engine,
    )
    expected = getattr(np, func)(counts, keepdims=True, dtype=dtype)
    assert actual.dtype == expected.dtype



import importlib
from contextlib import nullcontext

import numpy as np
import packaging.version
import pandas as pd
import pytest

pd_types = (pd.Index,)

try:
    import dask
    import dask.array as da

    dask_array_type = da.Array
except ImportError:
    dask_array_type = ()  # type: ignore[assignment, misc]


try:
    import xarray as xr

    xr_types = (xr.DataArray, xr.Dataset)
except ImportError:
    xr_types = ()  # type: ignore[assignment]


def _importorskip(modname, minversion=None):
    try:
        mod = importlib.import_module(modname)
        has = True
        if minversion is not None:
            if LooseVersion(mod.__version__) < LooseVersion(minversion):
                raise ImportError("Minimum version not satisfied")
    except ImportError:
        has = False
    func = pytest.mark.skipif(not has, reason=f"requires {modname}")
    return has, func


def LooseVersion(vstring):
    # Our development version is something like '0.10.9+aac7bfc'
    # This function just ignored the git commit id.
    vstring = vstring.split("+")[0]
    return packaging.version.Version(vstring)


has_cftime, requires_cftime = _importorskip("cftime")
has_cubed, requires_cubed = _importorskip("cubed")
has_dask, requires_dask = _importorskip("dask")
has_numba, requires_numba = _importorskip("numba")
has_numbagg, requires_numbagg = _importorskip("numbagg")
has_scipy, requires_scipy = _importorskip("scipy")
has_xarray, requires_xarray = _importorskip("xarray")


class CountingScheduler:
    """Simple dask scheduler counting the number of computes.

    Reference: https://stackoverflow.com/questions/53289286/"""

    def __init__(self, max_computes=0):
        self.total_computes = 0
        self.max_computes = max_computes

    def __call__(self, dsk, keys, **kwargs):
        self.total_computes += 1
        if self.total_computes > self.max_computes:
            raise RuntimeError(f"Too many computes. Total: {self.total_computes} > max: {self.max_computes}.")
        return dask.get(dsk, keys, **kwargs)


def raise_if_dask_computes(max_computes=0):
    # return a dummy context manager so that this can be used for non-dask objects
    if not has_dask:
        return nullcontext()
    scheduler = CountingScheduler(max_computes)
    return dask.config.set(scheduler=scheduler)


def assert_equal(a, b, tolerance=None):
    __tracebackhide__ = True

    if isinstance(a, list):
        a = np.array(a)
    if isinstance(b, list):
        b = np.array(b)

    if isinstance(a, pd_types) or isinstance(b, pd_types):
        pd.testing.assert_index_equal(a, b)
        return
    if has_xarray and isinstance(a, xr_types) or isinstance(b, xr_types):
        xr.testing.assert_identical(a, b)
        return

    if tolerance is None:
        if np.issubdtype(a.dtype, np.float64) | np.issubdtype(b.dtype, np.float64):
            tolerance = {"atol": 1e-18, "rtol": 1e-15}
        else:
            tolerance = {}

    # Always run the numpy comparison first, so that we get nice error messages with dask.
    # sometimes it's nice to see values and shapes
    # rather than being dropped into some file in dask
    if a.dtype != b.dtype:
        raise AssertionError(f"a and b have different dtypes: (a: {a.dtype}, b: {b.dtype})")

    if has_dask:
        a_eager = a.compute() if isinstance(a, dask_array_type) else a
        b_eager = b.compute() if isinstance(b, dask_array_type) else b
    else:
        a_eager, b_eager = a, b

    if a.dtype.kind in "SUMmO":
        np.testing.assert_equal(a_eager, b_eager)
    else:
        np.testing.assert_allclose(a_eager, b_eager, equal_nan=True, **tolerance)

    if has_dask and isinstance(a, dask_array_type) or isinstance(b, dask_array_type):
        # does some validation of the dask graph
        dask_assert_eq(a, b, equal_nan=True)


def assert_equal_tuple(a, b):
    """assert_equal for .blocks indexing tuples"""
    assert len(a) == len(b)

    for a_, b_ in zip(a, b):
        assert type(a_) is type(b_)
        if isinstance(a_, np.ndarray):
            np.testing.assert_array_equal(a_, b_)
        else:
            assert a_ == b_


SCIPY_STATS_FUNCS = ("mode", "nanmode")
BLOCKWISE_FUNCS = ("median", "nanmedian", "quantile", "nanquantile") + SCIPY_STATS_FUNCS
ALL_FUNCS = (
    "sum",
    "nansum",
    "argmax",
    "nanfirst",
    "nanargmax",
    "prod",
    "nanprod",
    "mean",
    "nanmean",
    "var",
    "nanvar",
    "std",
    "nanstd",
    "max",
    "nanmax",
    "min",
    "nanmin",
    "argmin",
    "nanargmin",
    "any",
    "all",
    "nanlast",
    "median",
    "nanmedian",
    "quantile",
    "nanquantile",
) + tuple(SCIPY_STATS_FUNCS)


def dask_assert_eq(
    a,
    b,
    check_shape=True,
    check_graph=True,
    check_meta=True,
    check_chunks=True,
    check_ndim=True,
    check_type=True,
    check_dtype=True,
    equal_nan=True,
    scheduler="sync",
    **kwargs,
):
    """dask.array.utils.assert_eq modified to skip value checks. Their code is buggy for some dtypes.
    We just check values through numpy and care about validating the graph in this function."""
    from dask.array.utils import _get_dt_meta_computed

    a_original = a
    b_original = b

    if isinstance(a, list | int | float):
        a = np.array(a)
    if isinstance(b, list | int | float):
        b = np.array(b)

    a, adt, a_meta, a_computed = _get_dt_meta_computed(
        a,
        check_shape=check_shape,
        check_graph=check_graph,
        check_chunks=check_chunks,
        check_ndim=check_ndim,
        scheduler=scheduler,
    )
    b, bdt, b_meta, b_computed = _get_dt_meta_computed(
        b,
        check_shape=check_shape,
        check_graph=check_graph,
        check_chunks=check_chunks,
        check_ndim=check_ndim,
        scheduler=scheduler,
    )

    if check_type:
        _a = a if a.shape else a.item()
        _b = b if b.shape else b.item()
        assert type(_a) is type(_b), f"a and b have different types (a: {type(_a)}, b: {type(_b)})"
    if check_meta:
        if hasattr(a, "_meta") and hasattr(b, "_meta"):
            dask_assert_eq(a._meta, b._meta)
        if hasattr(a_original, "_meta"):
            msg = (
                f"compute()-ing 'a' changes its number of dimensions "
                f"(before: {a_original._meta.ndim}, after: {a.ndim})"
            )
            assert a_original._meta.ndim == a.ndim, msg
            if a_meta is not None:
                msg = (
                    f"compute()-ing 'a' changes its type "
                    f"(before: {type(a_original._meta)}, after: {type(a_meta)})"
                )
                assert type(a_original._meta) is type(a_meta), msg
                if not (np.isscalar(a_meta) or np.isscalar(a_computed)):
                    msg = (
                        f"compute()-ing 'a' results in a different type than implied by its metadata "
                        f"(meta: {type(a_meta)}, computed: {type(a_computed)})"
                    )
                    assert type(a_meta) is type(a_computed), msg
        if hasattr(b_original, "_meta"):
            msg = (
                f"compute()-ing 'b' changes its number of dimensions "
                f"(before: {b_original._meta.ndim}, after: {b.ndim})"
            )
            assert b_original._meta.ndim == b.ndim, msg
            if b_meta is not None:
                msg = (
                    f"compute()-ing 'b' changes its type "
                    f"(before: {type(b_original._meta)}, after: {type(b_meta)})"
                )
                assert type(b_original._meta) is type(b_meta), msg
                if not (np.isscalar(b_meta) or np.isscalar(b_computed)):
                    msg = (
                        f"compute()-ing 'b' results in a different type than implied by its metadata "
                        f"(meta: {type(b_meta)}, computed: {type(b_computed)})"
                    )
                    assert type(b_meta) is type(b_computed), msg



import pytest
from hypothesis import HealthCheck, Verbosity, settings

from . import requires_numbagg

settings.register_profile(
    "ci",
    max_examples=1000,
    deadline=None,
    suppress_health_check=[HealthCheck.filter_too_much, HealthCheck.too_slow],
)
settings.register_profile(
    "default",
    max_examples=300,
    deadline=500,
    suppress_health_check=[HealthCheck.filter_too_much, HealthCheck.too_slow],
    verbosity=Verbosity.verbose,
)
settings.load_profile("default")


@pytest.fixture(
    scope="module",
    params=[
        "flox",
        "numpy",
        # pytest.param("numba", marks=requires_numba),
        pytest.param("numbagg", marks=requires_numbagg),
    ],
)
def engine(request):
    return request.param


@pytest.fixture(
    scope="module",
    params=[
        "flox",
        "numpy",
        pytest.param("numbagg", marks=requires_numbagg),
    ],
)
def engine_no_numba(request):
    return request.param



from dataclasses import dataclass

from .types import DaskArray, Graph


@dataclass
class ArrayLayer:
    name: str
    layer: Graph
    chunks: tuple[tuple[int, ...], ...]

    def to_array(self, dep: DaskArray) -> DaskArray:
        from dask.array import Array
        from dask.highlevelgraph import HighLevelGraph

        graph = HighLevelGraph.from_collections(self.name, self.layer, dependencies=[dep])
        return Array(graph, self.name, self.chunks, meta=dep._meta)



#!/usr/bin/env python
# flake8: noqa
"""Top-level module for flox ."""

from . import cache
from .aggregations import Aggregation, Scan  # noqa
from .core import (
    groupby_reduce,
    groupby_scan,
    rechunk_for_blockwise,
    rechunk_for_cohorts,
)  # noqa


def _get_version():
    __version__ = "999"
    try:
        from ._version import __version__
    except ImportError:
        pass
    return __version__


__version__ = _get_version()



import builtins
import math
from functools import partial
from itertools import product
from numbers import Integral

from dask import config
from dask.blockwise import lol_tuples
from toolz import partition_all

from .lib import ArrayLayer
from .types import Graph


# _tree_reduce and partial_reduce are copied from dask.array.reductions
# They have been modified to work purely with graphs, and without creating new Array layers
# in the graph. The `block_index` kwarg is new and avoids a concatenation by simply setting the right
# key initially
def _tree_reduce(
    x: ArrayLayer,
    *,
    name: str,
    out_dsk: Graph,
    aggregate,
    axis: tuple[int, ...],
    block_index: int,
    split_every=None,
    combine=None,
):
    # Normalize split_every
    split_every = split_every or config.get("split_every", 4)
    if isinstance(split_every, dict):
        split_every = {k: split_every.get(k, 2) for k in axis}
    elif isinstance(split_every, Integral):
        n = builtins.max(int(split_every ** (1 / (len(axis) or 1))), 2)
        split_every = dict.fromkeys(axis, n)
    else:
        raise ValueError("split_every must be a int or a dict")

    numblocks = tuple(len(c) for c in x.chunks)
    out_chunks = x.chunks

    # Reduce across intermediates
    depth = 1
    for i, n in enumerate(numblocks):
        if i in split_every and split_every[i] != 1:
            depth = int(builtins.max(depth, math.ceil(math.log(n, split_every[i]))))
    func = partial(combine or aggregate, axis=axis)

    agg_dep_name = x.name
    for level in range(depth - 1):
        newname = name + f"-{block_index}-partial-{level}"
        out_dsk, out_chunks = partial_reduce(
            func,
            out_dsk,
            chunks=out_chunks,
            split_every=split_every,
            name=newname,
            dep_name=agg_dep_name,
            axis=axis,
        )
        agg_dep_name = newname
    func = partial(aggregate, axis=axis)
    return partial_reduce(
        func,
        out_dsk,
        chunks=out_chunks,
        split_every=split_every,
        name=name,
        dep_name=agg_dep_name,
        axis=axis,
        block_index=block_index,
    )


def partial_reduce(
    func,
    dsk,
    *,
    chunks: tuple[tuple[int, ...], ...],
    name: str,
    dep_name: str,
    split_every: dict[int, int],
    axis: tuple[int, ...],
    block_index: int | None = None,
):
    numblocks = tuple(len(c) for c in chunks)
    ndim = len(numblocks)
    parts = [list(partition_all(split_every.get(i, 1), range(n))) for (i, n) in enumerate(numblocks)]
    keys = product(*map(range, map(len, parts)))
    out_chunks = [
        tuple(1 for p in partition_all(split_every[i], c)) if i in split_every else c
        for (i, c) in enumerate(chunks)
    ]
    for k, p in zip(keys, product(*parts)):
        free = {i: j[0] for (i, j) in enumerate(p) if len(j) == 1 and i not in split_every}
        dummy = dict(i for i in enumerate(p) if i[0] in split_every)
        g = lol_tuples((dep_name,), range(ndim), free, dummy)
        assert dep_name != name
        if block_index is not None:
            k = (*k[:-1], block_index)
        dsk[(name,) + k] = (func, g)
    return dsk, out_chunks



from functools import partial

import numbagg
import numbagg.grouped
import numpy as np
from packaging.version import Version

NUMBAGG_SUPPORTS_DDOF = Version(numbagg.__version__) >= Version("0.7.0")

DEFAULT_FILL_VALUE = {
    "nansum": 0,
    "nanmean": np.nan,
    "nanvar": np.nan,
    "nanstd": np.nan,
    "nanmin": np.nan,
    "nanmax": np.nan,
    "nanany": False,
    "nanall": False,
    "nansum_of_squares": 0,
    "nanprod": 1,
    "nancount": 0,
    "nanargmax": np.nan,
    "nanargmin": np.nan,
    "nanfirst": np.nan,
    "nanlast": np.nan,
}

CAST_TO = {
    # "nansum": {np.bool_: np.int64},
    "nanmean": {np.int_: np.float64},
    "nanvar": {np.int_: np.float64},
    "nanstd": {np.int_: np.float64},
    "nanfirst": {np.datetime64: np.int64, np.timedelta64: np.int64},
    "nanlast": {np.datetime64: np.int64, np.timedelta64: np.int64},
}


FILLNA = {"nansum": 0, "nanprod": 1}


def _numbagg_wrapper(
    group_idx,
    array,
    *,
    func,
    axis=-1,
    size=None,
    fill_value=None,
    dtype=None,
    **kwargs,
):
    cast_to = CAST_TO.get(func, None)
    if cast_to:
        for from_, to_ in cast_to.items():
            if np.issubdtype(array.dtype, from_):
                array = array.astype(to_, copy=False)

    func_ = getattr(numbagg.grouped, f"group_{func}")

    result = func_(
        array,
        group_idx,
        axis=axis,
        num_labels=size,
        **kwargs,
        # The following are unsupported
        # fill_value=fill_value,
        # dtype=dtype,
    ).astype(dtype, copy=False)

    return result


def nanvar(group_idx, array, *, axis=-1, size=None, fill_value=None, dtype=None, ddof=0):
    kwargs = {}
    if NUMBAGG_SUPPORTS_DDOF:
        kwargs["ddof"] = ddof
    elif ddof != 1:
        raise ValueError("Need numbagg >= v0.7.0 to support ddof != 1")
    return _numbagg_wrapper(
        group_idx,
        array,
        axis=axis,
        size=size,
        func="nanvar",
        **kwargs,
        # fill_value=fill_value,
        # dtype=dtype,
    )


def nanstd(group_idx, array, *, axis=-1, size=None, fill_value=None, dtype=None, ddof=0):
    kwargs = {}
    if NUMBAGG_SUPPORTS_DDOF:
        kwargs["ddof"] = ddof
    elif ddof != 1:
        raise ValueError("Need numbagg >= v0.7.0 to support ddof != 1")
    return _numbagg_wrapper(
        group_idx,
        array,
        axis=axis,
        size=size,
        func="nanstd",
        **kwargs,
        # fill_value=fill_value,
        # dtype=dtype,
    )


def nanlen(group_idx, array, *, axis=-1, size=None, fill_value=None, dtype=None):
    if array.dtype.kind in "US":
        array = np.broadcast_to(np.array([1]), array.shape)
    return _numbagg_wrapper(
        group_idx,
        array,
        axis=axis,
        size=size,
        func="nancount",
        # fill_value=fill_value,
        # dtype=dtype,
    )


nansum = partial(_numbagg_wrapper, func="nansum")
nanmean = partial(_numbagg_wrapper, func="nanmean")
nanprod = partial(_numbagg_wrapper, func="nanprod")
nansum_of_squares = partial(_numbagg_wrapper, func="nansum_of_squares")
nanprod = partial(_numbagg_wrapper, func="nanprod")
nanfirst = partial(_numbagg_wrapper, func="nanfirst")
nanlast = partial(_numbagg_wrapper, func="nanlast")
# nanargmax = partial(_numbagg_wrapper, func="nanargmax)
# nanargmin = partial(_numbagg_wrapper, func="nanargmin)
nanmax = partial(_numbagg_wrapper, func="nanmax")
nanmin = partial(_numbagg_wrapper, func="nanmin")
any = partial(_numbagg_wrapper, func="nanany")
all = partial(_numbagg_wrapper, func="nanall")

# sum = nansum
# mean = nanmean
# sum_of_squares = nansum_of_squares



from typing import Any, TypeAlias

try:
    import cubed.Array as CubedArray
except ImportError:
    CubedArray = Any

try:
    import dask.array.Array as DaskArray
    from dask.typing import Graph
except ImportError:
    DaskArray = Any
    Graph: TypeAlias = Any  # type: ignore[no-redef,misc]



from __future__ import annotations

from collections.abc import Callable
from typing import Any

import cftime
import dask
import hypothesis.extra.numpy as npst
import hypothesis.strategies as st
import numpy as np

from . import ALL_FUNCS, SCIPY_STATS_FUNCS

Chunks = tuple[tuple[int, ...], ...]

calendars = st.sampled_from(
    [
        "standard",
        "gregorian",
        "proleptic_gregorian",
        "noleap",
        "365_day",
        "360_day",
        "julian",
        "all_leap",
        "366_day",
    ]
)


@st.composite
def units(draw, *, calendar: str) -> str:
    choices = ["days", "hours", "minutes", "seconds", "milliseconds", "microseconds"]
    if calendar == "360_day":
        choices += ["months"]
    elif calendar == "noleap":
        choices += ["common_years"]
    time_units = draw(st.sampled_from(choices))

    dt = draw(st.datetimes())
    year, month, day = dt.year, dt.month, dt.day
    if calendar == "360_day":
        day = min(day, 30)
    if calendar in ["360_day", "365_day", "noleap"] and month == 2 and day == 29:
        day = 28

    return f"{time_units} since {year}-{month}-{day}"


@st.composite
def cftime_arrays(
    draw: st.DrawFn,
    *,
    shape: st.SearchStrategy[tuple[int, ...]] = npst.array_shapes(),
    calendars: st.SearchStrategy[str] = calendars,
    elements: dict[str, Any] | None = None,
) -> np.ndarray[Any, Any]:
    if elements is None:
        elements = {}
    elements.setdefault("min_value", -10_000)
    elements.setdefault("max_value", 10_000)
    cal = draw(calendars)
    values = draw(npst.arrays(dtype=np.int64, shape=shape, elements=elements))
    unit = draw(units(calendar=cal))
    return cftime.num2date(values, units=unit, calendar=cal)


numeric_dtypes = (
    npst.integer_dtypes(endianness="=")
    | npst.unsigned_integer_dtypes(endianness="=")
    | npst.floating_dtypes(endianness="=", sizes=(32, 64))
    # TODO: add complex here not in supported_dtypes
)
numeric_like_dtypes = (
    npst.boolean_dtypes()
    | numeric_dtypes
    | npst.datetime64_dtypes(endianness="=")
    | npst.timedelta64_dtypes(endianness="=")
)
supported_dtypes = (
    numeric_like_dtypes
    | npst.unicode_string_dtypes(endianness="=")
    | npst.complex_number_dtypes(endianness="=")
)
by_dtype_st = supported_dtypes

NON_NUMPY_FUNCS = [
    "first",
    "last",
    "nanfirst",
    "nanlast",
    "count",
    "any",
    "all",
] + list(SCIPY_STATS_FUNCS)
SKIPPED_FUNCS = ["var", "std", "nanvar", "nanstd"]

func_st = st.sampled_from([f for f in ALL_FUNCS if f not in NON_NUMPY_FUNCS and f not in SKIPPED_FUNCS])
numeric_arrays = npst.arrays(
    elements={"allow_subnormal": False}, shape=npst.array_shapes(), dtype=numeric_dtypes
)
numeric_like_arrays = npst.arrays(
    elements={"allow_subnormal": False}, shape=npst.array_shapes(), dtype=numeric_like_dtypes
)
all_arrays = (
    npst.arrays(
        elements={"allow_subnormal": False},
        shape=npst.array_shapes(),
        dtype=numeric_like_dtypes,
    )
    | cftime_arrays()
)


def by_arrays(
    shape: st.SearchStrategy[tuple[int, ...]], *, elements: dict[str, Any] | None = None
) -> st.SearchStrategy[np.ndarray[Any, Any]]:
    if elements is None:
        elements = {}
    elements.setdefault("alphabet", st.characters(exclude_categories=["C"]))
    return st.one_of(
        npst.arrays(
            dtype=npst.integer_dtypes(endianness="=") | npst.unicode_string_dtypes(endianness="="),
            shape=shape,
            elements=elements,
        ),
        cftime_arrays(shape=shape, elements=elements),
    )


@st.composite
def chunks(draw: st.DrawFn, *, shape: tuple[int, ...]) -> Chunks:
    chunks = []
    for size in shape:
        if size > 1:
            nchunks = draw(st.integers(min_value=1, max_value=size - 1))
            dividers = sorted(
                set(draw(st.integers(min_value=1, max_value=size - 1)) for _ in range(nchunks - 1))
            )
            chunks.append(tuple(a - b for a, b in zip(dividers + [size], [0] + dividers)))
        else:
            chunks.append((1,))
    return tuple(chunks)


@st.composite
def chunked_arrays(
    draw: st.DrawFn,
    *,
    chunks: Callable[..., st.SearchStrategy[Chunks]] = chunks,
    arrays=all_arrays,
    from_array: Callable = dask.array.from_array,
) -> dask.array.Array:
    array = draw(arrays)
    chunks = draw(chunks(shape=array.shape))

    if array.dtype.kind in "cf":
        nan_idx = draw(
            st.lists(
                st.integers(min_value=0, max_value=array.shape[-1] - 1),
                max_size=array.shape[-1] - 1,
                unique=True,
            )
        )
        if nan_idx:
            array[..., nan_idx] = np.nan

    return from_array(array, chunks=chunks)
